procedure T_Assignments is
   procedure Type_Test is separate;
   procedure Sliding is separate;
   procedure Groupable_Repeated is separate;
   procedure Access_Duplication is separate;
begin
   null;
end T_Assignments;
