procedure T_Array_Declarations is
   procedure Indexes     is separate;
   procedure Dimensions  is separate;
   procedure Component   is separate;
   procedure Index_Types is separate;   
begin
   null;
end T_Array_Declarations;
