with Ada.Exceptions;
procedure T_exception_propagation is
   procedure X_Declaration_Parameter is separate;
   procedure X_Interface             is separate;
   procedure X_Task                  is separate;
   procedure X_Local_Exception       is separate;
begin
   null;
end T_exception_propagation;

