procedure T_With_Clauses is
   procedure Plain   is separate;
   procedure Tgen    is separate;
   procedure Taspect is separate;
   procedure Limited_And_Private is separate;
begin
   null;
end T_With_Clauses;
