procedure T_Unit_Pattern is
   procedure Test_Single_Tagged_Type    is separate;
   procedure Test_Tagged_Type_Hierarchy is separate;
   procedure Test_Context_Clauses_Order is separate;
   procedure Test_Declarations_Order    is separate;
begin
   null;
end T_Unit_Pattern;
