procedure t_object_declarations is
   procedure Test_Min_Integer_Span  is separate;
   procedure Test_Not_Required_Type is separate;
   procedure Test_Type              is separate;
   procedure Test_Volatile_Address  is separate;
begin
   null;
end T_Object_Declarations;
