with Ada.Strings.Fixed;            -- others
with X_Dependencies.Child;         -- others, public child
with X_Dependencies.Priv;          -- others, private child
package body X_Dependencies is     -- more than 2, less than 10, all direct
begin
   null;
end X_Dependencies;
