package body X_Style_Casing.Child_1.Child_2 is   -- Casing_Identifier

   procedure P2 is   -- Casing_Identifier
   begin
      null;
   end P2;           -- Casing_Identifier

end X_Style_Casing.Child_1.Child_2;             -- Casing_Identifier x2
