with Ada.Strings.Fixed;             -- others
with Ada, Calendar, Ada.Calendar;   -- others (x3)
package X_Dependencies.Child is     -- more than 2, less than 10, all direct, not beween 5 and 10
   procedure P;
end X_Dependencies.Child;
