with Text_IO;                -- forbidden
with Ada.Text_IO;
with Ada.Numerics;           -- others
with Ada, Ada, Ada;          -- others (x3)ema
package X_Dependencies is    -- more than 2, less than 10, all direct, not between 5 and 10
   pragma Elaborate_Body;
end X_Dependencies;
