procedure T_Derivations is
   procedure From is separate;
   procedure Max_Parents is separate;
   procedure Indicator is separate;
   procedure Max_Depth is separate;
begin
   null;
end T_Derivations;
