--
-- Author: Rosen
-- Changed: 1
-- Changed: 2



----
--*
--**
--***
-- At least one
-- At least one
--end

with X_Header_Comments;
procedure T_Header_Comments is
   procedure Sep is separate;
begin
   null;
end T_Header_Comments;
