procedure T_Max_Line_Length is
   -- This line is 50 characters long ------------
   -- This line is 51 characters long -------------
   -- This line is 61 characters long -----------------------
   --## Rule line off max_line_length ## This line is long but inhibited
begin
   null;
end T_Max_Line_Length;

