procedure T_simplifiable_expressions is
   procedure Test_Range       is separate;
   procedure Test_Logical     is separate;
   procedure Test_Parentheses is separate;
   procedure Test_Conversion  is separate;
   procedure Test_Membership  is separate;
begin
   null;
end T_simplifiable_expressions;
