with X_Dependencies.Child;       -- others
with X_Dependencies_Ren;         -- others
procedure T_Dependencies is      -- less than 10, all direct, not between 5 and 10
begin
   null;
end T_Dependencies;
