procedure T_Comments is
   procedure Test_Pattern       is separate;
   procedure Test_Position      is separate;
   procedure Test_Terminating   is separate;
   procedure Test_Unnamed_Begin is separate;
begin  -- T_Comments
   null;
end T_Comments;

