procedure T_Expressions is
   procedure Test_Aggregate            is separate;
   procedure Test_Attributes           is separate;
   procedure Test_Call                 is separate;
   procedure Test_Conversions          is separate;
   procedure Test_Dereference          is separate;
   procedure Test_Fixed_Multiplying_Op is separate;
   procedure Test_Logical_Operators    is separate;
   procedure Test_Membership           is separate;
   procedure Test_Mixed                is separate;
   procedure Test_Others               is separate;
   procedure Test_Prefixed_Operator    is separate;
   procedure Test_Range                is separate;
   procedure Test_Real_Equality        is separate;
   procedure Test_Slice                is separate;
   procedure Test_2012                 is separate;
begin
   null;
end T_Expressions;
