separate (T_Comments)
procedure Test_Position is
begin  --Test_Position
                                       -- column 40        OK
                                        -- column 41       search
                                         -- column 42      search
                                          -- column 43     check
   null;
end Test_Position;
